import grpc
import json
import time
import copy
import typing


from deepdiff import DeepDiff
from perxis.collections import collections_pb2_grpc, collections_pb2
from perxis.roles import roles_pb2_grpc, roles_pb2
from perxis.common import common_pb2
from perxis.clients import clients_pb2_grpc, clients_pb2
from perxis.environments import environments_pb2_grpc, environments_pb2


class ExtensionSetup:
    def __init__(
            self,
            collections_service: collections_pb2_grpc.CollectionsStub,
            environments_service: environments_pb2_grpc.EnvironmentsStub,
            roles_service: roles_pb2_grpc.RolesStub,
            clients_service: clients_pb2_grpc.ClientsStub,
    ):
        self.collections = []
        self.clients = []
        self.roles = []

        self.roles_service = roles_service
        self.clients_service = clients_service
        self.collections_service = collections_service
        self.environments_service = environments_service

        self.__max_attempts_count = 5
        self.__sleep_time = 1

    def add_collection(self, collection: collections_pb2.Collection):
        self.collections.append(collection)

    def set_collections(self, collections: list[collections_pb2.Collection]):
        self.collections = collections

    def add_role(self, role: roles_pb2.Role):
        self.roles.append(role)

    def set_roles(self, roles: list[roles_pb2.Role]):
        self.roles = roles

    def add_client(self, client: clients_pb2.Client):
        self.clients.append(client)

    def set_clients(self, clients: list[clients_pb2.Client]):
        self.clients = clients

    # Работа с ролями
    def __remove_roles(self, space_id: str) -> list[str]:
        errors_list: list[str] = []

        for role in self.roles:
            try:
                self.roles_service.Delete.with_call(
                    roles_pb2.DeleteRequest(
                        space_id=space_id, role_id=role.id
                    )
                )
            except grpc.RpcError as e:
                # Если роли не существует считать это ошибкой не надо
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить роль {role.id}, {e.details()}")

        return errors_list

    def __check_roles(self, space_id: str) -> list[str]:
        errors_list = []

        for role in self.roles:
            try:
                self.roles_service.Get.with_call(
                    roles_pb2.GetRequest(space_id=space_id, role_id=role.id)
                )
            except grpc.RpcError as e:
                errors_list.append(f"Не удалось получить роль {role.id}, ошибка {e.details()}")

        return errors_list

    def __update_roles(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []

        for local_role in self.roles:
            try:
                get_response, state = self.roles_service.Get.with_call(
                    roles_pb2.GetRequest(
                        space_id=space_id,
                        role_id=local_role.id
                    )
                )

                role = get_response.role
            except grpc.RpcError as e:
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось получить роль {local_role.id}, {e.details()}")
                    continue

                role = None

            cloned_role = copy.deepcopy(local_role)
            cloned_role.space_id = space_id

            if role:
                if not role.environments:
                    role.environments[:] = [env_id]
                elif env_id not in role.environments:
                    role.environments.append(env_id)

                cloned_role.environments[:] = role.environments

                # Произвести мерж правил доступа
                for exist_rule in role.rules:
                    was_found = False

                    for local_rule in cloned_role.rules:
                        # Если запись с правами доступа на коллекцию найдена - правила нужно совместить
                        if local_rule.collection_id == exist_rule.collection_id:
                            was_found = True

                            local_rule.actions[:] = list(set(list(local_rule.actions) + list(exist_rule.actions)))

                            break

                    # Если правило для коллекций не было найдено - его нужно добавить
                    if not was_found:
                        cloned_role.rules.append(common_pb2.Rule(
                            collection_id=exist_rule.collection_id,
                            actions=list(exist_rule.actions)
                        ))

                try:
                    self.roles_service.Update.with_call(
                        roles_pb2.UpdateRequest(
                            role=cloned_role
                        )
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось обновить роль {local_role.id}, {e.details()}")
            else:
                if env_id not in cloned_role.environments:
                    cloned_role.environments.append(env_id)

                try:
                    cloned_role = copy.deepcopy(local_role)
                    cloned_role.space_id = space_id

                    response, _ = self.roles_service.Create.with_call(
                        roles_pb2.CreateRequest(
                            role=cloned_role
                        ),
                    )
                except grpc.RpcError as e:
                    # На этапе install считается что ролей __нет__. При install с указанием force роли предварительно
                    # удаляются
                    errors_list.append(f"Не удалось создать роль {local_role.id}, {e.details()}")

        return errors_list

    # Работа с клиентами
    def __check_clients(self, space_id: str) -> list[str]:
        errors_list = []

        for client in self.clients:
            try:
                self.clients_service.Get.with_call(
                    clients_pb2.GetRequest(space_id=space_id, id=client.id)
                )
            except grpc.RpcError as e:
                errors_list.append(f"Не удалось получить клиент {client.id}, ошибка {e.details()}")

        return errors_list

    def __remove_clients(self, space_id: str) -> list[str]:
        errors_list: list[str] = []

        for client in self.clients:
            try:
                self.clients_service.Delete.with_call(
                    clients_pb2.DeleteRequest(
                        space_id=space_id, id=client.id
                    )
                )
            except grpc.RpcError as e:
                # Отсутствие клиента ошибкой не считается
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить клиент {client.id}, {e.details()}")

        return errors_list

    def __update_clients(self, space_id: str) -> list[str]:
        errors_list = []

        for local_client in self.clients:
            try:
                # Перед обновлением клиента предварительно нужно получить текущую запись чтобы скопировать оттуда
                # токены. Иначе после обновления расширения перестанут работать все приложения которые использовали
                # токен клиента
                get_response, state = self.clients_service.Get.with_call(
                    clients_pb2.GetRequest(
                        space_id=space_id,
                        id=local_client.id
                    )
                )

                client = get_response.client
            except grpc.RpcError as e:
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось получить клиент {local_client.id}, {e.details()}")
                    continue

                client = None

            if client:
                try:
                    # Нужно чтобы у клиента каждый раз не слетали данные токенов при переустановке
                    # свойства oauth, api_key и tls должны браться из __созданного__ клиента
                    new_client = clients_pb2.Client(
                        id=local_client.id,
                        space_id=space_id,
                        name=local_client.name,
                        description=local_client.description,
                        disabled=client.disabled,
                        role_id=local_client.role_id,
                        oauth=client.oauth or local_client.oauth,
                        api_key=client.api_key or local_client.api_key,
                        tls=client.tls
                    )

                    self.clients_service.Update.with_call(
                        clients_pb2.UpdateRequest(
                            client=new_client
                        )
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось обновить клиент {local_client.id}, {e.details()}")
            else:
                try:
                    cloned_client = copy.deepcopy(local_client)
                    cloned_client.space_id = space_id

                    response, _ = self.clients_service.Create.with_call(
                        clients_pb2.CreateRequest(
                            client=cloned_client
                        ),
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось создать клиент {local_client.id}, {e.details()}")

        return errors_list

    # Работа с коллекциями
    def __remove_collections(self, space_id: str, env_id: str) -> list[str]:
        errors_list: list[str] = []

        for collection in self.collections:
            try:
                self.collections_service.Delete.with_call(
                    collections_pb2.DeleteRequest(
                        space_id=space_id, env_id=env_id, collection_id=collection.id
                    )
                )
            except grpc.RpcError as e:
                # Отсутствие коллекции это не ошибка
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить коллекцию {collection.id}, {e.details()}")

        return errors_list

    def __check_collections(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []

        for collection in self.collections:
            try:
                self.collections_service.Get.with_call(
                    collections_pb2.GetRequest(space_id=space_id, env_id=env_id, collection_id=collection.id)
                )
            except grpc.RpcError as e:
                errors_list.append(f"Не удалось получить коллекцию {collection.id}, ошибка {e.details()}")

        return errors_list

    def __update_collections(self, space_id: str, env_id: str) -> list[str]:
        """
            Метод __обновления__ коллекций. Миграция окружения требуется
            только в случае если одна или несколько схем коллекций изменялись. Алгоритм работы:
            1. Получить фактически существующую коллекцию из БД
            2. Обновить её в perxis
            3. Сравнить схему коллекции в расширении и в perxis
            4. Если схема изменена - обновить схему в perxis
            5. Если обновлялась хотя бы одна схема в perxis - запустить миграцию окружения
        """

        need_to_migrate_environment = False
        errors_list = []

        for local_collection in self.collections:
            try:
                # Необходимо получить текущую версию коллекции для того чтобы сравнить схемы
                get_response, state = self.collections_service.Get.with_call(
                    collections_pb2.GetRequest(space_id=space_id, env_id=env_id, collection_id=local_collection.id)
                )

                collection = get_response.collection
            except grpc.RpcError as e:
                collection = None

            cloned_collection = copy.deepcopy(local_collection)
            cloned_collection.space_id = space_id
            cloned_collection.env_id = env_id

            # Коллекция может быть не найдена в случае если она добавлена в новой версии
            if not collection:
                try:
                    create_response, state = self.collections_service.Create.with_call(
                        collections_pb2.CreateRequest(collection=cloned_collection)
                    )

                    collection = create_response.created
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось создать коллекцию {local_collection.id}, {e.details()}")

                    continue
            else:
                try:
                    cloned_collection = copy.deepcopy(local_collection)
                    cloned_collection.space_id = space_id
                    cloned_collection.env_id = env_id

                    self.collections_service.Update.with_call(
                        collections_pb2.UpdateRequest(collection=cloned_collection)
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось обновить коллекцию {local_collection.id}, {e.details()}")

                    continue

            diff = DeepDiff(
                json.loads(collection.schema or "{}"),
                json.loads(local_collection.schema or "{}"),
                ignore_numeric_type_changes=True,
                exclude_paths=["root['loaded']"]
            )
            if diff:
                need_to_migrate_environment = True

                set_schema_error_message = self.__set_collection_schema(
                    space_id, env_id, local_collection.id, local_collection.schema
                )
                if set_schema_error_message:
                    errors_list.append(set_schema_error_message)

        if need_to_migrate_environment:
            migrate_environment_error_message = self.__migrate_environment(space_id, env_id)
            if migrate_environment_error_message:
                errors_list.append(migrate_environment_error_message)

        return errors_list

    def __migrate_environment(self, space_id: str, env_id: str) -> typing.Optional[str]:
        # Так как perxis может не сразу выставить коллекции / окружению статус ready операцию необходимо выполнять
        # с попытками

        attempt = 0
        is_ok = False
        error_message = None

        while attempt <= self.__max_attempts_count and not is_ok:
            time.sleep(self.__sleep_time)

            try:
                self.environments_service.Migrate(environments_pb2.MigrateRequest(
                    space_id=space_id,
                    env_id=env_id
                ))

                is_ok = True
            except grpc.RpcError as e:
                # Если не удалось мигрировать окружение по любой причине кроме подготовки - это ошибка
                if "is preparing" not in e.details():
                    error_message = e.details()

                    # Для принудительного выхода из цикла
                    attempt = self.__max_attempts_count

            attempt += 1

        return error_message

    def __set_collection_schema(self, space_id: str, env_id: str, collection_id: str, schema: str) -> typing.Optional[str]:
        # Так как perxis может не сразу выставить коллекции / окружению статус ready операцию необходимо выполнять
        # с попытками

        attempt = 0
        is_ok = False
        error_message = None

        while attempt <= self.__max_attempts_count and not is_ok:
            time.sleep(self.__sleep_time)

            try:
                self.collections_service.SetSchema.with_call(
                    collections_pb2.SetSchemaRequest(
                        space_id=space_id,
                        env_id=env_id,
                        collection_id=collection_id,
                        schema=schema
                    )
                )

                is_ok = True
            except grpc.RpcError as e:
                # Если не удалось установить схему по любой причине кроме подготовки - это ошибка
                if "is preparing" not in e.details():
                    error_message = e.details()

                    # Для принудительного выхода из цикла
                    attempt = self.__max_attempts_count

            attempt += 1

        return error_message

    def install(self, space_id: str, env_id: str, use_force: bool) -> list[str]:
        errors = []

        if use_force:
            errors += self.__remove_collections(space_id, env_id)
            errors += self.__remove_clients(space_id)
            errors += self.__remove_roles(space_id)

        errors += self.__update_collections(space_id, env_id)
        errors += self.__update_roles(space_id, env_id)
        errors += self.__update_clients(space_id)

        return errors

    def update(self, space_id: str, env_id: str, use_force: bool) -> list[str]:
        errors = []

        # В случае обновление расширения с флагом force нужно предварительно удалить все сущности.
        if use_force:
            errors += self.__remove_clients(space_id)
            errors += self.__remove_roles(space_id)
            errors += self.__remove_collections(space_id, env_id)

        errors += self.__update_collections(space_id, env_id)
        errors += self.__update_roles(space_id, env_id)
        errors += self.__update_clients(space_id)

        return errors

    def check(self, space_id: str, env_id: str) -> list[str]:
        errors = self.__check_collections(space_id, env_id)
        errors += self.__check_roles(space_id)
        errors += self.__check_clients(space_id)

        return errors

    def uninstall(self, space_id: str, env_id: str, use_remove: bool) -> list[str]:
        errors = []

        if use_remove:
            errors += self.__remove_collections(space_id, env_id)
            errors += self.__remove_clients(space_id)
            errors += self.__remove_roles(space_id)

        return errors
